local cpml = require "cpml"
local iqm  = require "iqm"

local models, shader, vignette, noise, start = {}, false, false, false, 0
local using_gles = select(1, love.graphics.getRendererInfo()) == "OpenGL ES"

function love.load()
	-- Monkey-patch the love API - you can use the lib without this by
	-- omitting "true", but it makes things a bit easier to use. The
	-- second argument disables automatic transforms.
	require("love3d").import(true, false)

	-- Usually you'd define this stuff in some external file.
	local data = {
		{
			model = "assets/aesa.iqm",
			use_fresnel = true,
			textures = {
				body = "assets/skin.png",
				dress = "assets/dress.png"
			}
		}, {
			model = "assets/ground.iqm",
			textures = {
				ground = "assets/ground.png"
			}
		}
	}

	-- Load all the textures and set filtering. If you don't set the
	-- filtering everything looks bad at a distance.
	local flags = {
		mipmaps = true
	}

	for _, params in ipairs(data) do
		local t = love.timer.getTime()
		local model = iqm.load(params.model)

		model.fresnel = params.use_fresnel and 1 or 0
		model.textures = {}
		for material, path in pairs(params.textures) do
			model.textures[material] = love.graphics.newImage(path, flags)
			model.textures[material]:setFilter("linear", "linear", 16)
		end

		table.insert(models, model)
	end

	-- Simple shader with an edge highlight.
	shader = love.graphics.newShader("shader.glsl")

	if not using_gles then
		-- Things to pretty the demo up just a little
		-- We mix a little noise into the vignette to break up banding.
		vignette = love.graphics.newShader("vignette.glsl")
		noise    = love.graphics.newImage("assets/noise.png", flags)
		noise:setWrap("repeat", "repeat")
	end
	start = love.timer.getTime()

	love.graphics.setBackgroundColor(241, 244, 249)
end

function love.draw()
	love.graphics.push("all")

	-- Love doesn't use the depth buffer, so we need to set it up & clear it.
	-- We need depth testing so we don't have to depth-sort opaque objects.
	love.graphics.clearDepth()
	love.graphics.setDepthTest("less")
	love.graphics.setColor(255, 255, 255, 255)

	-- Disable blending.
	love.graphics.setBlendMode("replace", "premultiplied")

	-- Now we transform & update the shader params...
	-- World transform. love.graphics.{translate, rotate, scale} are hooked
	local speed = 0.25
	love.graphics.setShader(shader)

	-- View transform (camera)
	local vp = cpml.mat4():look_at(
		cpml.vec3(0, 5, 1.5), -- look from 0, 5, 1.5...
		cpml.vec3(0, 0, 1),   -- to 0, 0, 1...
		cpml.vec3.unit_z      -- and +z is up.
	)

	-- Projection
	local w, h = love.graphics.getDimensions()
	vp = vp * cpml.mat4():perspective(25, w/h, 0.01, 1000)
	shader:send("u_viewProjection", vp:to_vec4s())

	-- Model transform
	local m = cpml.mat4():rotate((love.timer.getTime() - start) * speed + math.pi / 2, cpml.vec3(0, 0, -1))
	shader:send("u_model", m:to_vec4s())

	-- You can draw the whole model as one mesh (just don't set draw
	-- range), but if you draw this way you can assign a different
	-- shader/textures per-mesh. (but at the slight expense of more
	-- draw calls - so do what's best for you)
	for _, model in ipairs(models) do
		shader:sendInt("use_fresnel", model.fresnel and 1 or 0)
		for _, buffer in ipairs(model) do
			local texture = model.textures[buffer.material]
			model.mesh:setTexture(texture) -- nil is OK
			model.mesh:setDrawRange(buffer.first, buffer.last)
			love.graphics.draw(model.mesh)
		end
	end

	-- Make sure to reset things back to how they were, or you'll
	-- have a bad time. Don't forget to reset blending, too!
	love.graphics.setDepthTest()
	love.graphics.setShader()
	love.graphics.pop()

	-- Disable vignette on GLES devices (it's slow).
	if not using_gles then
		love.graphics.setShader(vignette)
		love.graphics.setBlendMode("multiply")
		love.graphics.setColor(255, 255, 255, 255)
		love.graphics.draw(noise, 0, 0, 0, w / noise:getWidth(), h / noise:getHeight())
	end

	-- Reset blending, shader.
	love.graphics.setBlendMode("alpha")
	love.graphics.setShader()

	love.graphics.setColor(0, 0, 0, 255)
	love.graphics.print(string.format("FPS: %0.2f (%0.3fms)", love.timer.getFPS(), love.timer.getAverageDelta()), 5, 5)
	love.graphics.print("Press ESCAPE to quit.", 5, 25)
end

function love.keypressed(k, s, r)
	if k == "escape" then love.event.quit() end
end
